//
// Author:  Carlo Barth
// Date:    2017/10/21
//
// Purpose: A high-quality, artistic render of quantum dots (QDs) near a
//          photonic crystal surface, interacting with a leaky mode.
//          The leaky mode is represented by a color-coded volume render.
//          The QDs have an emission proportional to the local field energy
//          density, denoted by a glowing aura.
//
// Details: The data is based on a finite element method solve of the
//          photonic crystal on glass. The glass substrate is omitted in the
//          render. See the related paper describing most of the experimental
//          and numerical methods and results:
//
//              http://aip.scitation.org/doi/10.1063/1.4995229
//

// Imports
//--------------------------------------------------------------------------
#include "colors.inc"
#include "textures.inc"
#include "glass.inc"
#include "metals.inc"
#include "rand.inc"
#include "shapes.inc"


//==========================================================================
//                  SECTION: Data read-in, parameters and settings
//==========================================================================

// Global Settings
//--------------------------------------------------------------------------
#version 3.7;
global_settings{ 
    assumed_gamma 1.0 
}

// String that identifies the field mode and quantum dot data folder
//--------------------------------------------------------------------------
#ifndef ( Which_mode )
    #declare Which_mode  = "mode1";
#end

// Filenames
#declare N_QDs = 1180;
#declare FN_qd_coords =concat("mode_data/",Which_mode, "/qd_choice_50_1180.dat");
#declare FN_aura = concat("mode_data/", Which_mode, "/efield_aura_50.df3");
#declare FN_extent = concat("mode_data/", Which_mode, "/extent.dat");
#declare FN_fR = concat("mode_data/", Which_mode, "/efield_R.df3");
#declare FN_fG = concat("mode_data/", Which_mode, "/efield_G.df3");
#declare FN_fB = concat("mode_data/", Which_mode, "/efield_B.df3");

// Real geometry parameters in nanometer
#declare a_real = 600.; // lattice constant
#declare d_real = 367.; // hole diameter
#declare h_real = 116.; // phc height
#declare h_sup_real = 250.; // superspace height
#declare r_qds_real = 2.8; // radius quantum dots (QD)
#declare r_qds_aura = 12; // maximum radius factor for QD aura

// Quantum dot positions and intensity data
//--------------------------------------------------------------------------
#declare QD_coords = array[N_QDs];
#declare QD_intensities = array[N_QDs];
#declare Index = 0;
#fopen QDFile FN_qd_coords  read
#while (defined(QDFile))
    #read (QDFile, X, Y, Z, QD_intensity)
    #declare QD_coords[Index] = <Y/a_real, (Z-h_sup_real)/a_real, X/a_real>;
    #declare QD_intensities[Index] = QD_intensity;
    #declare Index = Index + 1;
#end

// Field and aura density data extent
//--------------------------------------------------------------------------
#declare Extent = array[2];
#declare Index = 0;
#fopen ExFile FN_extent read
#while (defined(ExFile))
    #read (ExFile, X, Y, Z)
    #declare Extent[Index] = <X, Y, Z>;
    #declare Index = Index + 1;
#end
#declare Exs = (Extent[1] - Extent[0])/a_real;
#declare Ext = Extent[0]/a_real;
#declare DF3_scale = <Exs.y, Exs.z, Exs.x>;
#declare DF3_translate = <Ext.y, Ext.z, Ext.x>;

// Global parameters
//--------------------------------------------------------------------------
// 0=off, 1=fast; 2=medium quality; 3=high qual.; 
// 4= medium qual.+ recursion_limit 2
#ifndef ( Radiosity_On )
    #declare Radiosity_On  = 0;
#end
#declare HDRlighting = true;
#declare Spot_on = true; // whether to use spot lights
#declare Area_light_on = true; // whether to use area light
#declare Area_light_brightness = 0.4; // area light brightness
#ifndef ( Camera_Type )
    // 1 = wide view
    // 2 = close-up
    // 3 = orthographic top-view
    // 4 = orthographic top-view far
    // 5 = orthographic top-view close
    #declare Camera_Type  = 1; 
#end

// Derived parameters and abbreviating expressions
#declare a = a_real/a_real;
#declare d = d_real/a_real;
#declare Edge_radius = d/15.; // hole edge rounding radius
#declare h = h_real/a_real;
#declare h_sup = h_sup_real/a_real;
#declare r_qds = r_qds_real/a_real;
#declare sqrt3 = 0.5*sqrt(3.); // for hexagonal lattice
#declare cx = a/sqrt3/4; // standard x coordinate of hexagon
#declare cy = 0.5*a; // standard y coordinate of hexagon
#declare Repeat_unit_cells = true; // whether to repeat the unit cell
#ifndef ( rows )
    #declare rows = 2; // (2) rows of the unit cell repetition pattern
#end
#ifndef ( cols )
    #declare cols = 1; // (1) columns of the unit cell repetition pattern
#end

// Parameters related to artistic depections, rather than real properties
//--------------------------------------------------------------------------
// Quantum dots
#declare QD_render_version = 1; // 2 fails for large number of QDs
#ifndef ( Show_QDs )
    #declare Show_QDs = true; // whether to render the quantum dots
#end

// Field/aura rendering
#declare QD_brightness = 8.; // brightness of the quantum dots
#declare r_aura = r_qds*r_qds_aura; // radius QD aura max (in diameters)
#ifndef ( Show_field )
    #declare Show_field = true; // whether to render the field or not
#end
#ifndef ( Show_auras )
    #declare Show_auras = true; // whether to render the auras or not
#end
//#declare Show_field = true; // whether to render the field or not
#declare Field_brightness = 2.; //2.; // brightness of the field render
#declare Field_interpolation = 1; // interpolation of the field/aura densities
#ifndef ( rows )
    #declare Field_samples = 1; // number of samples of the field/aura densities
#end
#declare Focal_blur = false; // whether to use focal blur

// Slicing option
#ifndef ( Slice )
    // whether to cut the PhC and Field along the xz-plane
    #declare Slice = false; 
#end

// Additional lights settings
#ifndef ( QD_lights )
    // whether to place an additional light source at each QD position 
    #declare QD_lights = false; 
#end
#declare QD_lights_brightness = 1.;

// Camera
//--------------------------------------------------------------------------
#declare Light_vert_shift = 1.; // Shifts the lights in +y-direction
camera {
    #switch (Camera_Type)
        // -----------------------------------------
        #case (1)
        orthographic angle 30
        location <0, 2, -3>
        look_at <0, h, 0>
        #break
        // -----------------------------------------
        #case (2)
        angle 30
        location <-cx*2.5, h*4, -cy*2.>
        look_at <cx*0.2, h/2, cy*.2>
        #break
        // -----------------------------------------
        #case (3)
        orthographic angle 50
        location < 0, 5, 0 > 
        look_at < 0, 0, 0 >
        #declare Light_vert_shift = 2.;
        #break
        // -----------------------------------------
        #case (4)
        orthographic angle 50
        location < 0 , 7, 0 > 
        look_at < 0, 0, 0 >
        #declare Light_vert_shift = 5.;
        #break
        // -----------------------------------------
        #case (5)
        orthographic angle 50
        location < 0 , 1.5, 0 > 
        look_at < 0, 0, 0 >
        #break
        // -----------------------------------------
        #case (6)
        orthographic angle 50
        location < 0 , h, -2*cy > 
        look_at < 0, h, 0 >
        #break
    #end // end of switch
    
    // Correct aspect for different image formats
    right x*image_width/image_height
    
    #if (Focal_blur)
        aperture 0.001 // [0...N] larger is narrower depth of field (blurrier)
        blur_samples 100 // number of rays per pixel for sampling
        focal_point < 0, h, 0 > // point that is in focus <X,Y,Z>
        confidence 0.95 // [0...<1] (smaller is less accurate)
        variance 1/200 // [0...1] (smaller is more accurate)
    #end
}


//==========================================================================
//                  SECTION: Illumination
//==========================================================================

// Radiosity
//--------------------------------------------------------------------------
#if (Radiosity_On > 0)

global_settings {
    #ifndef ( Rad_Quality )
        #declare Rad_Quality = Radiosity_On;
    #end

    //--------- radiosity settings -------------------------------- ///
    // from POV-Ray samples "scene templates/patio-radio.pov

    #switch (Rad_Quality)
    #case (1)
    radiosity {             // --- Settings 1 (fast) ---
        pretrace_start 0.08
        pretrace_end   0.02
        count 50
        error_bound 0.5
        recursion_limit 1
        media on
    }
    #break
    #case (2)
    radiosity {             // --- Settings 2 (medium quality) ---
        pretrace_start 0.08
        pretrace_end   0.01
        count 120
        error_bound 0.25
        recursion_limit 1
        media on
    }
    #break
    #case (3)
    radiosity {             // --- Settings 3 (high quality) ---
        pretrace_start 0.08
        pretrace_end   0.005
        count 400
        error_bound 0.1
        recursion_limit 1
        media on
    }
    #break
    #case (4)
    radiosity {             // --- Settings 4 (medium quality, recursion_limit)
        pretrace_start 0.08
        pretrace_end   0.005
        count 350
        error_bound 0.15
        recursion_limit 2
        media on
    }
    #break
    #case (5)
    radiosity {             // --- Settings 5 (high quality, recursion_limit )
        pretrace_start 0.08
        pretrace_end   0.005
        count 350
        error_bound 0.1
        recursion_limit 2
        media on
    }
    #break
    #end  // end of switch
    } // end of global settings

    #default{ finish {emission 0. diffuse 0.3 ambient 0.}}

//--------------------------------------------------------------------------
#else

    #default{ finish{ ambient 0. diffuse 0.9 }
    } // for intel computers

#end

// Light sources
//--------------------------------------------------------------------------

// A group of 3 spot lights
#if (Spot_on)
    light_source{ // greenish
        <1.,15*h*Light_vert_shift,-1> color rgb <.8,1,.8>/Light_vert_shift
        spotlight
        point_at<0.5,0,0>
        radius 60  // hotspot
        tightness 100
        falloff 90
    }

    light_source{ // blueish
        <-1.,15*h*Light_vert_shift,-1> color rgb <.8,.8,1>/Light_vert_shift
        spotlight
        point_at<-0.5,0,0>
        radius 60  // hotspot
        tightness 100
        falloff 90
    }

    light_source{  // "white"
        <0,15*h*Light_vert_shift,-1> color rgb <.8,.8,.8>/Light_vert_shift
        spotlight
        point_at<0.,0,-0.5>
        radius 60  // hotspot
        tightness 100
        falloff 90
    }
#end

#if (Area_light_on)
    // soft light (areal_light) ---------  An area light (creates soft shadows)
    // WARNING: This special light can significantly slow down rendering times!
    light_source { 
        <0,0,0>          
        color rgb <1,1,1>*Area_light_brightness/Light_vert_shift    // light's color 
        area_light            // kind of light source
        <-cx, 0, 0> <cx, 0, 0> // lights spread out across this distance (x * z)
        1, 10                // total number of lights in grid
        adaptive 2          // 0,1,2,3...
        jitter              // adds random softening of light
        //looks_like { object{Tube} }
        translate <0, 1.50*Light_vert_shift, 0>   // <x,y,z> position of light
    } //---------------- end of area_light
#end

//==========================================================================
//                  SECTION: Object description
//==========================================================================

// Textures
//--------------------------------------------------------------------------

// Material of the photonic crystal
#declare t_shiny_metal =
texture {
    pigment { 
        rgb <.55, .42, .41>
    }
    finish { 
        specular 0.2 roughness 0.5
        reflection 0.05
        brilliance 4
    }
}

// Macros
//--------------------------------------------------------------------------

// The quantum dot sphere macro (version 2) plus light source (optional)
#macro QuantumDot_sphere(Radius, Intensity, Origin)
    // The actual quantum dot
    sphere{ 
        <0,0,0>, Radius
        texture{ Glass2 } // end of texture
        no_reflection
        translate Origin
    }
    // Additional light source
    #if (QD_lights)
        light_source{ 
            Origin 
            color White*QD_lights_brightness/N_QDs
            fade_distance 2*r_aura
            fade_power 2
        }
    #end
#end // ------------------ end of macro

// Loop for triangular lattice
//--------------------------------------------------------------------------
#macro TriangularLoop ( Object, trans_vec )
    #local ii = -cols;
    #while ( ii <= cols )
        #local jj = -rows;
        #while ( jj <= rows )
            #if ( mod(ii,2) = 0 )
              #local hexshift = 0.;
            #else
              #local hexshift = 0.5*a;
            #end

            object{
              Object
              translate <ii*a*sqrt3, 0., jj*a+hexshift>
              translate trans_vec
            } 

            #local jj = jj+1;
        #end
        #local ii = ii+1;
    #end
#end // ------------------ end of macro


// Compound objects
//--------------------------------------------------------------------------

// Silicon membrane for complete PhC definition without unit cell
#declare geo_si_membrane = box { <-100,0,-100>, <100,h,100> }

// The two cylinders that make the hole
#declare geo_hole_cylinder = union {
    cylinder{<0, h-Edge_radius, 0>, <0, h+0.0001, 0>, d/2.+Edge_radius}
    cylinder { <0., -0.0001, 0.>, <0., h+0.0001, 0.>, d/2. }    
}

// The torus for the rounded edge of the hole
#declare geo_hole_torus = torus{ 
    d/2.+Edge_radius, Edge_radius 
    translate <0, h-Edge_radius, 0>
}

// All holes
#declare geo_all_hole_cylinders = union{
    TriangularLoop ( geo_hole_cylinder, <0,0,0> )
}

// All torus
#declare geo_all_torus = union{
    TriangularLoop ( geo_hole_torus, <0,0,0> )
}

#if (Slice)
    #declare phc = difference { union{
        difference{
            object{ geo_si_membrane }
            #if (Repeat_unit_cells)
                object{ geo_all_hole_cylinders }
            #else
                object{ geo_hole_cylinder }
            #end
        }
        #if (Repeat_unit_cells)
            object{ geo_all_torus }
        #else
            object{ geo_hole_torus }
        #end
    } // end of union
    box{ <-100,-0.1,0>, <100,h+0.1,-100> }
    } // end of intersection
    
    // The complete superspace volume
    #declare geo_superspace = difference {
                        box { <-100,-0.1,-100>, <100,h+h_sup,100> }
                        box { <-100,-0.2,0>, <100,h+h_sup+0.1,-100> }
    } // end of intersection
#else
    #declare phc = union{
        difference{
            object{ geo_si_membrane }
            #if (Repeat_unit_cells)
                object{ geo_all_hole_cylinders }
            #else
                object{ geo_hole_cylinder }
            #end
        }
        #if (Repeat_unit_cells)
            object{ geo_all_torus }
        #else
            object{ geo_hole_torus }
        #end
    } // end of union
    
    // The complete superspace volume
    #declare geo_superspace = box { <-10,-0.1,-10>, <10,h+h_sup,10> }
#end

// A single unit cell of the photonic crystal
//--------------------------------------------------------------------------
#declare PhC_object = object{
    phc
    texture{ t_shiny_metal }
    normal{ bozo .5
      scale <1,0.25,0.25>*1
      turbulence 2.
    }
}

// Union of the QDs
//--------------------------------------------------------------------------
#declare QDs_all = union {   
    #local ii = 0;
    #while ( ii < N_QDs )
        #local Int_i = QD_intensities[ii];
        #local Vec_i = QD_coords[ii];
        QuantumDot_sphere( r_qds, Int_i, Vec_i)
        #local ii = ii+1;
    #end
}

//==========================================================================
//                  SECTION: Objects in scene
//==========================================================================

// Superpace with quantum dot auras and field render
//--------------------------------------------------------------------------
object { 
    geo_superspace
    pigment { rgbt 1 } hollow
    interior {
        ior                 1.000
        caustics            0.000
        dispersion          1.000
        dispersion_samples  7.000
        fade_power          0.000
        fade_distance       0.000
        fade_color          rgb <0.000,0.000,0.000>
        
        //-----------------------------------------------------------
        // The field media: one for each color channel (R,G,B) ------
        //-----------------------------------------------------------
        #if (Show_field)
        // Red
        media {
            method     3
            intervals  10
            samples    Field_samples, 10*Field_samples
            confidence 0.900
            variance   0.008
            ratio      0.900
            absorption rgb <0,0,0>
            emission   rgb <1,0,0> * Field_brightness
            aa_threshold 0.050
            aa_level    4
            density {
             density_file df3 FN_fR
             interpolate Field_interpolation
            }
            // Set proper size and position
            scale DF3_scale
            translate DF3_translate
        }

        // Green
        media {
            method     3
            intervals  10
            samples    Field_samples, 10*Field_samples
            confidence 0.900
            variance   0.008
            ratio      0.900
            absorption rgb <0,0,0>
            emission   rgb <0,1,0> * Field_brightness
            aa_threshold 0.050
            aa_level    4
            density {
             density_file df3 FN_fG
             interpolate Field_interpolation
            }
            // Set proper size and position
            scale DF3_scale
            translate DF3_translate
        }

        // Blue
        media {
            method     3
            intervals  10
            samples    Field_samples, 10*Field_samples
            confidence 0.900
            variance   0.008
            ratio      0.900
            absorption rgb <0,0,0>
            emission   rgb <0,0,1> * Field_brightness
            aa_threshold 0.050
            aa_level    4
            density {
             density_file df3 FN_fB
             interpolate Field_interpolation
            }
            // Set proper size and position
            scale DF3_scale
            translate DF3_translate
        }
        #end
        //---------- End of field media -----------------------------
        
        //-----------------------------------------------------------
        // The quantum dot media: from aura df3 ---------------------
        //-----------------------------------------------------------
        #if (Show_auras)
        media {
            method     3
            intervals  10
            samples    Field_samples, 10*Field_samples
            confidence 0.900
            variance   0.008
            ratio      0.900
            absorption rgb <0,0,0>
            emission   rgb <1,1,1> * QD_brightness
            aa_threshold 0.050
            aa_level    4
            density {
             density_file df3 FN_aura
             interpolate Field_interpolation
            }
            // Set proper size and position
            scale DF3_scale
            translate DF3_translate
        }
        #end
        //---------- End of quantum dot media ------------------------
    }
}

// HDRI environment
//--------------------------------------------------------------------------
#if (HDRlighting)
    sky_sphere{
        pigment{
            image_map{
                hdr "hdr/Mt-Washington-Hotel-Lookout_Gauss_lesser.hdr"
                gamma 1.
                map_type 1
                interpolate 2
            }
        }
        rotate < 0,20,0>  // adapt it to your direction
    } // end sky_sphere

# else
    // Background (sky)
    sky_sphere{ 
        pigment{ 
            gradient <0,1,0>
            color_map{ [0   color rgb<1,1,1>         ]//White
                       [0.4 color rgb<0.14,0.14,0.56>]//~Navy
                       [0.6 color rgb<0.14,0.14,0.56>]//~Navy
                       [1.0 color rgb<1,1,1>         ]//White
                     }
            scale 2 }
    } // end of sky_sphere
#end

// Floor
//--------------------------------------------------------------------------
plane { 
    y-0.001, 0.
    texture{ Polished_Chrome }
    finish { reflection 0.2 ambient 0. }
    pigment { rgb <.05, .05, .1> }
    normal{ 
        bozo 0.1//bumps 0.03
        scale <1,0.25,0.25>*1
        turbulence 5.6
    }
}


// Photonic crystal
PhC_object

//  Quantum dot spheres
#if (Show_QDs)
    QDs_all
#end

